;;; find-cmd-widget.el --- Interactive find(1) frontend -*- lexical-binding: t -*-

;; Copyright (C) 2020 Free Software Foundation, Inc

;; Author: Michael Heerdegen <michael_heerdegen@web.de>
;; Maintainer: Michael Heerdegen <michael_heerdegen@web.de>
;; Created: 10 Jun 2019
;; Keywords: convenience
;; Compatibility: GNU Emacs 26
;; Version: 0.1
;; Package-Requires: ((emacs "26"))


;; This file is not part of GNU Emacs.

;; GNU Emacs is free software: you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation, either version 3 of the License, or
;; (at your option) any later version.

;; GNU Emacs is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with GNU Emacs.  If not, see <http://www.gnu.org/licenses/>.

;;; Commentary:
;;;
;;; This package offers a feature complete interactive find(1)
;;; frontend for Emacs: find(1) expressions can be build using widgets
;;; and completion having explaining annotations that makes find(1)
;;; features more discoverable.  Checks for some common pitfalls
;;; interactively and warns you.  The built find command can be viewed
;;; and copied from a text field and, more importantly, can be run in
;;; Emacs with C-c C-c.
;;; Available actions include running as command, viewing
;;; matches in dired and running `grep'.  To get started just do
;;; M-x find-cmd-widget.
;;;
;;; At top-level and inside any `not' pattern there is a button named
;;; `and` automatically added.  `and`-combination is the default, and
;;; this button is added for convenience - if you later find "oh no, I
;;; needed the `or` combiner", you can toggle `and` and `or` by
;;; clicking on that button without the need to start from scratch.
;;; Of course you can also explicitly add any combiners yourself.
;;;
;;; This package also allows to add your own patterns (see
;;; `find-cmd-widget-additional-patterns' and
;;; `find-cmd-widget-user-patterns') and mix them with built-in ones.
;;; This is useful to match, for example, media files based on a
;;; specified list of extensions and such things, without the need to
;;; specify the extensions every time.
;;;
;;; There is no particular setup required to start.  Maybe you want to
;;; have a look at the options defined in "find-dired.el" like
;;; find-ls-subdir-switches' for example.


;;; Code:

;;;; Requirements

(require 'cl-lib)
(require 'subr-x)
(require 'widget)
(require 'find-cmd)
(require 'cus-edit)
(eval-when-compile (require 'wid-edit))

(defgroup find-cmd-widget nil
  "Expression based search and replace for Emacs Lisp."
  :group 'find-dired)

;;;; Variables

(defvar find-cmd-widget-find-arg-global-options-arglist
  ;; elements are (symbol make-widget-fun category annotation hint-string)
  ;;               0      1               2        3          4
  `((depth                 find-cmd-widget-make-choice-widget         "Global Options"
                           "Process directory contents before the directory itself")
    (maxdepth              find-cmd-widget-make-n-widget              "Global Options"
                           "Minimum level of descend")
    (mindepth              find-cmd-widget-make-n-widget              "Global Options"
                           "Maximum level of descend")
    (mount                 find-cmd-widget-make-choice-widget         "Global Options"
                           "Don't descend directories on other filesystems (alternate name for xdev)")
    (noleaf                find-cmd-widget-make-choice-widget         "Global Options"
                           "\
Don't assume dirs contain 2 fewer subdirs than their hard link count.")
    (ignore_readdir_race   find-cmd-widget-make-choice-widget         "Global Options"
                           "Ignore certain errors")
    (noignore_readdir_race find-cmd-widget-make-choice-widget         "Global Options"
                           "Turns off the effect of -ignore_readdir_race")
    (xdev                  find-cmd-widget-make-choice-widget         "Global Options"
                           "Don't descend directories on other filesystems")
    (warn                  find-cmd-widget-make-choice-widget         "Global Options"
                           "Turn `find` warning messages on")
    (nowarn                find-cmd-widget-make-choice-widget         "Global Options"
                           "Turn `find` warning messages off")))

(defcustom find-cmd-widget-add-to-command-history t
  "Doc..."
  :type 'boolean)

(defvar find-cmd-widget-find-arg-alist
  ;; elements are (symbol make-widget-fun category annotation hint-string)
  ;;               0      1               2        3          4
  `(
    ;; generic tests
    (amin       find-cmd-widget-make-time-minutes-widget   "Tests"
                "Minutes since file last accessed")
    (anewer     find-cmd-widget-make-xnewer-widget         "Tests"
                "Last accessed more recently than a specified file was modified")
    (atime      find-cmd-widget-make-time-days-widget      "Tests"
                "Days since file last accessed (rounded down; see also `daystart`)")
    (cmin       find-cmd-widget-make-time-minutes-widget   "Tests"
                "Minutes since last status change")
    (cnewer     find-cmd-widget-make-xnewer-widget         "Tests"
                "Status changed more recently than a specified file was modified")
    (ctime      find-cmd-widget-make-time-days-widget      "Tests"
                "Days since last status change (rounded down; see also `daystart`)")
    (empty      find-cmd-widget-make-choice-widget         "Tests"
                "Empty regular file or directory")
    (executable find-cmd-widget-make-choice-widget         "Tests"
                "File is executable/ Dir is searchable")
    (false      find-cmd-widget-make-choice-widget         "Tests"
                "Always false")
    ;; "follow" is deprecated
    (fstype     find-cmd-widget-make-editable-field-widget "Tests"
                "File is on a filesystem of specified type")
    (gid        ,(lambda (name hint)
                   (find-cmd-widget-make-signed-n-widget
                    name hint "numeric group ID" '=))
                "Tests" "File's numeric group ID")
    (group      find-cmd-widget-make-editable-field-widget "Tests"
                "File belongs to specified group" "group name or numeric group ID")
    (ilname     find-cmd-widget-make-lname-widget          "Tests"
                "Case insensitive version of `lname`"
                "shell glob pattern; matched against the complete name sans dir")
    (iname      find-cmd-widget-make-editable-field-widget "Tests"
                "Like `name` but matching is case insensitive"
                "shell glob pattern; matched against the complete name sans dir")
    (inum       find-cmd-widget-make-editable-field-widget "Tests"
                "File has specified inode number" "Note: often `samefile` is simpler to use")
    (ipath      find-cmd-widget-make-editable-field-widget "Tests"
                "Like `path` but match is case insensitive"
                "shell glob pattern; matched against the complete path")
    (iregex     find-cmd-widget-make-editable-field-widget "Tests"
                "Like `regex` but matching is case insensitive"
                "regexp; matched against the complete path")
    ;; (iwholename (1) "tests") ; same as `ipath`
    (links      ,(lambda (name hint)
                   (find-cmd-widget-make-signed-n-widget name hint "links" '=))
                "Tests" "Number of hard links")
    (lname      find-cmd-widget-make-lname-widget          "Tests"
                "symlink whose target matches a shell pattern"
                "shell glob pattern; matched against the complete name (sans dir)")
    (mmin       find-cmd-widget-make-time-minutes-widget   "Tests"
                "Minutes since last data change")
    (mtime      find-cmd-widget-make-time-days-widget      "Tests"
                "Days since last data change (rounded down; see also `daystart`)")
    (name       find-cmd-widget-make-editable-field-widget "Tests"
                "Match shell pattern against base of file name"
                "shell glob pattern; matched against the complete name (sans dir)")
    (newer      find-cmd-widget-make-xnewer-widget         "Tests"
                "Last modified more recently than a specified file")
    (newerXY    ,(lambda (name _hint)
                   `(find-cmd-widget-newerXY-widget :menu-tag ,(symbol-name name)))
                                                           "Tests"
                "Compare certain timestamp with a timestamp of specified file")
    (nogroup    find-cmd-widget-make-choice-widget         "Tests"
                "No group corresponds to file's numeric group ID")
    (nouser     find-cmd-widget-make-choice-widget         "Tests"
                "No user corresponds to file's numeric user ID")
    (path       find-cmd-widget-make-editable-field-widget "Tests"
                "Match shell pattern against complete file name"
                "shell glob pattern; matched against the complete path")
    (perm       find-cmd-widget-make-perm-widget           "Tests"
                "File permissions (octal or symbolic)")
    (readable   find-cmd-widget-make-choice-widget         "Tests" "Readable file")
    (regex      find-cmd-widget-make-editable-field-widget "Tests"
                "Match regexp against file path"
                "regexp; matched against the complete path")
    (samefile   ,(lambda (name hint)
                   `(find-cmd-widget-file-widget
                     :value-get ,(lambda (w) `(,name ,(expand-file-name (widget-value-value-get w))))
                     :tag ,(symbol-name name)
                     :format ,(concat (symbol-name name)
                                      " as %[file%]: %v"
                                      (when hint (format "          [%s]" hint))
                                      "\n")))
                "Tests" "File refers to same inode as file specified")
    (size       find-cmd-widget-make-size-widget           "Tests"
                "File size")
    (true       find-cmd-widget-make-choice-widget         "Tests" "Always true")
    (type       find-cmd-widget-make-type-widget           "Tests" "File Type")
    (uid        ,(lambda (name hint)
                   (find-cmd-widget-make-signed-n-widget
                    name hint "numeric user ID" '=))
                "Tests" "File's numeric user ID")
    (used       find-cmd-widget-make-time-days-widget      "Tests"
                "Days since file last accessed after its last status change")
    (user       find-cmd-widget-make-editable-field-widget "Tests"
                "File owned by specified user" "user name or numeric user ID")
    ;; (wholename  (1) "tests") ; same as `path`
    (writable   find-cmd-widget-make-choice-widget         "Tests" "Writable file")
    (xtype      find-cmd-widget-make-type-widget           "Tests" "File or link target type")

    ;; normal options (always true)
    (daystart              find-cmd-widget-make-choice-widget         "Options"
                           "Measure times of following options from beginning of today")
    (regextype             find-cmd-widget-make-editable-field-widget "Options"
                           "Change regex syntax for later tests")

    (and   find-cmd-widget-make-combiner-widget "Combiners"
           "Logical 'and' operator")
    (not   find-cmd-widget-make-combiner-widget-w/operator "Combiners"
           "Logical 'not' operator")
    (or    find-cmd-widget-make-combiner-widget "Combiners"
           "Logical 'or' operator")
    (prune find-cmd-widget-make-choice-widget   "Combiners"
           ;; FIXME: add warnings (prune is ignored under certain circumstances
           ;; e.g. when combined with depth)
           "True; if the file is a directory, do not descend into it")

    ;; Actions

    (quit    find-cmd-widget-make-choice-widget         "Actions"
             "Exit immediately")

    ;; Skip these for now; user can use dired
    ;; (delete  find-cmd-widget-make-choice-widget         "Actions")
    ;; (print0  find-cmd-widget-make-choice-widget         "Actions")
    ;; (printf  find-cmd-widget-make-editable-field-widget "Actions")
    ;; (fprintf find-cmd-widget-make-editable-field-widget "Actions")
    ;; (print   find-cmd-widget-make-choice-widget         "Actions")
    ;; (fprint0 find-cmd-widget-make-editable-field-widget "Actions")
    ;; (fprint  find-cmd-widget-make-editable-field-widget "Actions")
    ;; (ls      find-cmd-widget-make-choice-widget         "Actions")
    ;; (fls     find-cmd-widget-make-editable-field-widget "Actions")

    ;; these need to be terminated with a ;
    ;; (exec    find-cmd-widget-make-editable-field-widget "Actions")
    ;; (ok      find-cmd-widget-make-editable-field-widget "Actions")
    ;; (execdir find-cmd-widget-make-editable-field-widget "Actions")
    ;; (okdir   find-cmd-widget-make-editable-field-widget "Actions")
    )
  "Doc...")

;; FIXME: how to get this handled correctly by find-cmd and find-dired?
(defvar find-cmd-widget-find-global-option-arglist
  '(;; switches
    (L find-cmd-widget-make-choice-widget "Switches"
       "Follow symbolic links")
    (P find-cmd-widget-make-choice-widget "Switches"
       "Never follow symbolic links (default)")
    (H find-cmd-widget-make-choice-widget "Switches"
       "Do not follow symlinks except while processing command line arguments"))
  "Doc...")

(defvar find-cmd-widget-defaults '())

(defvar-local find-cmd-widget-main-widget nil)
(defvar-local find-cmd-widget-preview-widget-1 nil)
(defvar-local find-cmd-widget-preview-widget-2 nil)
(defvar-local find-cmd-widget-warning-widget nil)

(defvar-local find-cmd-widget-current-global-option nil)
(defvar-local find-cmd-widget-current-global-options nil)
(defvar-local find-cmd-widget-warnings '())

(defvar find-cmd-widget-newerXY-X-list
  '(("specified time"     "t")
    ("access time"        "a")
    ("birth time"         "B")
    ("status change time" "c")
    ("modification time"  "m")))

(defvar find-cmd-widget-find-constituents
  `((prune . (0)) (warn . (0)) (nowarn . (0))
    ,@(let ((xs (mapcar #'cadr find-cmd-widget-newerXY-X-list)))
        (cl-mapcan (lambda (part) (mapcar (lambda (y) `(,(intern (concat part y)) . (1))) xs))
                   (mapcar (lambda (x) (concat "newer" x)) xs))))
  "Additional ...")

;;;; Additional predefined patterns, like for matching media files
;;;; etc.

(defun find-cmd-widget-make-extension-regexp (extensions &optional match-whole)
  (concat (when match-whole ".*") "\\.\\(" (mapconcat #'regexp-quote extensions "\\|") "\\)"))

(defvar find-cmd-widget-regex-media-file
  (find-cmd-widget-make-extension-regexp
   '("m4a" "m3u" "ogv" "flac" "mp2" "wma" "wav" "avi" "opus"
     "mp4" "mp3" "m4v" "3gp" "mpg" "mpeg"
     "qt" "mov" "moov" "rv" "fli" "flc" "flv" "mkv" "mng" "rm" "vro"
     "asf" "asx" "wmv" "ogm" "vob" "ogg" "mts" "webm" "ts" "divx")))

(defvar find-cmd-widget-regex-image-file
  (find-cmd-widget-make-extension-regexp
   '("jpg" "jpeg" "gif" "tiff" "png" "bmp" "psd" "xmp" "xbm" "eim" "pnm" "xpm" "webp" "svg")))

(defvar find-cmd-widget-additional-patterns
  `((media-file
     "Audio and Video Files [by file extension]"
     (iregex ,(concat ".*" find-cmd-widget-regex-media-file)))
    (image-file
     "Image Files [by file extension]"
     (iregex ,(concat ".*" find-cmd-widget-regex-image-file)))
    (grep
     "Files that contain matches for specified grep(1) call"
     ,(lambda (name _description)
        (let ((name-string (symbol-name name)))
          `(find-cmd-widget-editable-field
            :menu-tag ,name-string
            ;; Yes, we could massively extend that...
            :format ,(concat "grep -q -e" " %v" (format "  (%s)" "regexp") "\n")
            :value-get ,(lambda (w)
                          `(and
                            (type "f")
                            (exec ; final ";" added automatically by `find-cmd'
                             ,(concat grep-program " " "-q" ;; find-grep-options
                                      " -e " (shell-quote-argument (widget-field-value-get w))
		                      " " (shell-quote-argument "{}")))))))))))

(defcustom find-cmd-widget-user-patterns '()
  "Doc...
See `find-cmd-widget-additional-patterns' for some examples.

Some more examples:

\\='(regular-file \"\\
Regular file (symlink behavior depends on -L or -follow options)\"
               (type \"f\"))

\\='(directory    \"\\
Directory (symlink behavior depends on -L or -follow options)\"
               (type \"d\"))

Pattern names can be used as \"definition\" to define another
name as an alias for that pattern:

\\='(owner \"Synonymous with `user`\" user)

In the most low-level form, the definition is a function that returns
a widget specification, e.g.

`(mctime
  \"Or\\='d ctime + mtime\"
  ,(lambda (name _descr)
     (let ((w (find-cmd-widget-make-signed-n-widget name nil \"days\" '<))
           (vget (lambda (w)
                   (let ((wv (apply #\\='concat
                                    (widget-editable-list-value-get w))))
                     `(or (ctime ,wv) (mtime ,(copy-sequence wv)))))))
       (setf (plist-get (cdr w) :value-get) vget)
       (setf (plist-get (cdr w) :format) \"Days since last file change:\\n%v\")
       w)))

(let ((format (lambda (name-string)
                (concat name-string
                        \"\\
: %v  (glob pattern; will be wrapped in \\\"*...*\\\")\" \"\\n\"))))
  `(*iname*
    \"iname with implicit \\\"*...*\\\" wrapping\"
    ,(lambda (name _descr)
       (let ((name-string (symbol-name name)))
         `(find-cmd-widget-editable-field
           :menu-tag ,name-string
           :format ,(funcall format name-string)
           :value-get
           ,(lambda (w)
              `(iname ,(format \"*%s*\" (widget-field-value-get w)))))))))

or

`(ipath+iname
  \"Handy combination of iname, ipath and file type\"
  ,(lambda (_name _descr)
     `(group
       :tag \"inp\"
       :args ((find-cmd-widget-editable-field
               :menu-tag \"iname\" :format \"iname %v  in \" :value \"**\")
              (find-cmd-widget-editable-field
               :menu-tag \"ipath\" :format \"ipath %v  of \" :value \"**\")
              (find-cmd-widget-editable-field
               :menu-tag \"type\" :format \"type %v\\n\"      :value \"f,d,l\"))
       :format \"%v\"
       :value-get ,(lambda (w)
                     (let ((results
                            (seq-remove
                             (pcase-lambda (`(,_ ,s))
                               (and (stringp s)
                                    (string-match-p
                                     (rx bos (* (any space ?*)) eos)
                                     s)))
                             (widget-editable-list-value-get w))))
                       (pcase results
                         ((pred cdr-safe) `(and ,@results))
                         ((pred identity) (car results))
                         (_ (list))))))))"
  :type 'sexp)

;;;; Helpers

;;;###autoload
(defmacro find-cmd-widget-with-additional-find-constituents (&rest body)
  (declare (debug t))
  `(let ((find-constituents (append find-cmd-widget-find-constituents find-constituents)))
     ,@body))

(defun find-cmd-widget--just-true (_) t)

(defun find-cmd-widget--warn-arg-not-implemented (arg)
  (when arg (warn "Argument not implemented")))

(defun find-cmd-widget--parse-widget-sexp (sexp)
  (cadr sexp))

(defun find-cmd-widget-create-find-cmd (sexp)
  `(find-cmd-widget-with-additional-find-constituents
    (find-cmd ,@(mapcar (lambda (x) `',x) (find-cmd-widget--parse-widget-sexp (cdr sexp))))))

(defun find-cmd-widget-find-call-get-dir+exps (sexp)
  (list (car sexp)
        (find-cmd-widget-with-additional-find-constituents
         (mapconcat #'find-to-string (find-cmd-widget--parse-widget-sexp (cdr sexp)) ""))))

(defun find-cmd-widget--build-dirs-string (dirs &optional no-shell-quote)
  (mapconcat (if no-shell-quote #'identity #'shell-quote-argument)
             (mapcar #'expand-file-name dirs) " "))

(defun find-cmd-widget-create-find-call (sexp)
  ;; We do that manually because `find-cmd' doesn't yet handle multiple
  ;; search paths
  (let* ((dirs+exps (find-cmd-widget-find-call-get-dir+exps sexp))
         (dirs (car dirs+exps)))
    (concat find-program
            " "
            (find-to-string find-cmd-widget-current-global-option)
            (if dirs (concat (find-cmd-widget--build-dirs-string dirs)) ".")
            " "
            (cadr dirs+exps))))

(defvar find-cmd-widget--widget-choose--widget-alist)
(defvar find-cmd-widget--widget-choose--use-categories)
(defun find-cmd-widget-menu-make-choice-action (widget-alist &optional use-categories)
  (lambda (&rest args)
    (cl-letf (((symbol-function #'widget-choose)
               #'find-cmd-widget--widget-choose))
      (let ((find-cmd-widget--widget-choose--widget-alist widget-alist)
            (find-cmd-widget--widget-choose--use-categories use-categories))
        (apply #'widget-choice-action args)))))

(defun find-cmd-widget--widget-choose (title items &optional event)
  (let ((widget-menu-max-size 100)
        (widget-alist find-cmd-widget--widget-choose--widget-alist))
    (if event
        ;; mouse click
        (let* (chosen-item
               (make-item
                (lambda (i)
                  (let* ((s (car i))
                         (e (assoc (intern s) widget-alist)))
                    (vector s (lambda () (interactive) (setq chosen-item (cdr i)))
                            :help (nth 3 e))))))
          (popup-menu
           `(,title
             (keymap
              ,@(and find-cmd-widget--widget-choose--use-categories
                     `(("By category"
                        ,@(mapcar
                           (pcase-lambda (`(,cat . ,items))
                             `(,cat ,@(mapcar make-item items)))
                           (seq-group-by
                            (lambda (item)
                              (nth 2 (assoc (intern (car item)) widget-alist)))
                            items)))))
              ,@(mapcar make-item (cl-sort items #'string< :key #'car))))
           event)
          chosen-item)
      (setq items (cl-remove-if 'stringp items))
      (let* ((max-item-length (apply #'max
                                     (mapcar (lambda (e) (length (symbol-name (car e))))
                                             widget-alist)))
             (max-cat-length  (apply #'max
                                     (mapcar #'length (mapcar #'caddr widget-alist))))
             (completions-format 'vertical)
             (val (completing-read
                   (concat title ": ")
                   (lambda (string pred action)
                     (if (eq action 'metadata)
	                 `(metadata
                           (display-sort-function . ,#'identity)
                           (annotation-function
                            .
                            ,(lambda (s)
                               (when-let ((entry (assoc (intern-soft s)
                                                        widget-alist)))
                                 (let ((category (nth 2 entry))
                                       (description (nth 3 entry)))
                                   (and (or category description)
                                        (concat (make-string (+ 1 (- max-item-length (length s))) ?\ )
                                                (format "[%s]" category)
                                                (make-string
                                                 (+ 1 (- max-cat-length (length category))) ?\ )
                                                (if description description ""))))))))
                       (complete-with-action action items string pred)))
                   nil t)))
	(and (stringp val)
	     (let ((try (try-completion val items)))
	       (when (stringp try)
	         (setq val try))
               (let ((res (cdr (assoc val items))))
                 (if-let ((alias (plist-get (cdr res) :alias-for)))
                     (cdr (assoc (symbol-name alias) items))
                   res))))))))

(defun find-cmd-widget--value-get-1 (command)
  (lambda (w) (list command (widget-field-value-get w))))

(defalias 'find-cmd-widget--get-arg
  (pcase-lambda ((and `(,name ,(and (pred functionp) make-widget) . ,rest)
                      (let hint (nth 2 rest))
                      (let widget (funcall make-widget name hint))))
    (cons (car widget)
          (append (alist-get name find-cmd-widget-defaults)
                  (cdr widget)))))

(defun find-cmd-widget-warn-maybe ()
  (when find-cmd-widget-warnings
    (message "There are warnings (see end of this buffer) but trying anyway...")
    (sit-for 1.5)))

(defun find-cmd-widget-find-action (&rest _)
  (interactive)
  (find-cmd-widget-warn-maybe)
  (when find-cmd-widget-add-to-command-history
    (push `(async-shell-command ,(widget-value find-cmd-widget-preview-widget-2))
          command-history))
  (async-shell-command (widget-value find-cmd-widget-preview-widget-2)))

(defun find-cmd-widget-find-dired-action--1 ()
  (let ((args (find-cmd-widget-find-call-get-dir+exps (widget-value find-cmd-widget-main-widget))))
    (when (cdr (car args))
      (user-error "find-dired currently supports only one path, sorry"))
    (find-cmd-widget-warn-maybe)
    (setf (car args) (if (car args) (find-cmd-widget--build-dirs-string (car args) 'no-quote) "."))
    args))

(defun find-cmd-widget-find-dired-action (&rest _)
  "Run constructed find(1) command with `find-dired'."
  (interactive)
  (let ((args (find-cmd-widget-find-dired-action--1))
        ;; Hack (but works...)
        (find-program
         (concat find-program " " (find-to-string find-cmd-widget-current-global-option))))
    (when find-cmd-widget-add-to-command-history
      (push `((lambda (&rest args)
                (interactive)
                (cl-letf (((default-value 'find-program) ,find-program))
                  (apply #'find-dired args)))
              ,@args)
            command-history))
    (cl-letf (((default-value 'find-program) find-program))
      ;; Bug in Emacs: when ARGS contain a global option, like in
      ;; ("." "-maxdepth 1 \\( -iname \\*abc\\* \\) ")
      ;; `find-dired' will wrap the global options into another \( ... \)
      ;; which causes `find` to complain that there is an argument before the
      ;; global option.  Can be suppressed with -nowarn.
      (apply #'find-dired args))))

(defun find-cmd-widget-make-grep-command-default-function (find-command grep-regexp remote?)
  (concat find-command
          ;; FIXME: this doesn't work in all environments - but it's hard to get
          ;; something out of `grep-compute-defaults'
          " -exec " grep-command grep-regexp " "
          (shell-quote-argument "{}" remote?) " "
          (shell-quote-argument "+" remote?)))

(defvar find-cmd-widget-make-grep-command-function #'find-cmd-widget-make-grep-command-default-function)

(defun find-cmd-widget-find-grep-action (grep-regexp)
  "Run constructed find(1) command using `grep-find'.
This requires to specify a grep(1) pattern in the in the
corresponding text field at the end of the \"*Widget Find*\"
buffer."
  (interactive)
  (require 'grep)
  (grep-compute-defaults)
  (let ((display-buffer-alist '(("" . (display-buffer-same-window)))))
    (grep-find
     (funcall find-cmd-widget-make-grep-command-function
              (concat find-program " " (find-to-string find-cmd-widget-current-global-option)
                      (mapconcat #'identity (find-cmd-widget-find-dired-action--1) " "))
              grep-regexp
              (file-remote-p default-directory)))))


;;;; Widget definitions

(define-widget 'find-cmd-preview 'item
  "Doc..."
  :format "%v " :size 0 ;see widget-field-value-create and the additional newline
  :value-face 'highlight
  :value-create
  (lambda (w)
    (widget-field-value-create w)
    (pcase-let ((`(,start . ,end) (widget-get w :field-overlay)))
      (put-text-property start end 'read-only t)))
  :value-set
  ;; widget-default-value-set deletes all undo information
  (lambda (widget value)
    (let ((buffer-undo-list))
      (widget-default-value-set widget value))))

(defvar find-cmd-widget-sign-args
  (mapcar (pcase-lambda (`(,sign ,descr ,val))
            (cons sign
                  `(choice-item :tag ,descr
                                :format "%[%t%] "
                                :value-inline ,(lambda (_w) `(,val)))))
          '((< "less than" "-") (> "more than" "+") (= "exactly" ""))))

(define-widget 'find-cmd-widget-sign 'menu-choice
  "Doc..."
  :format "%v\n" ;Unfortunately widgets break in diverse ways without the trailing \n
  :args (mapcar #'cdr find-cmd-widget-sign-args))

(defun find-cmd-widget-make-choice-widget (name hint)
  (find-cmd-widget--warn-arg-not-implemented hint)
  `(choice-item :tag ,(symbol-name name)
                :format ,(format "%%[%%t%%]          [%s]\n"
                                 (nth 3 (assoc name (append find-cmd-widget-find-global-option-arglist
                                                            find-cmd-widget-find-arg-global-options-arglist
                                                            find-cmd-widget-find-arg-alist))))
                :value-get ,(lambda (_w) (list name))))

(defun find-cmd-widget-make-user-pattern-widget (name _hint)
  (let ((entry (assoc name (append find-cmd-widget-user-patterns find-cmd-widget-additional-patterns))))
    (if (functionp (nth 2 entry))
        (funcall (nth 2 entry) (nth 0 entry) (nth 1 entry))
      `(choice-item :tag ,(symbol-name name)
                    :format ,(format "%%[%%t%%]          [%s]\n" (nth 1 entry))
                    :value-get ,(lambda (_w) (nth 2 entry))
                    ,@(when (symbolp (nth 2 entry)) `(:alias-for ,(nth 2 entry)))))))

(defvar find-cmd-widget-grep-widget)

(defun find-cmd-widget-default-action ()
  "Run constructed find(1) command.

Run the constructed find command using `find-dired' unless the
text field after the \"grep\" button is not empty - in this case
run `grep-find' using that grep regexp.

This is the default command executed when hitting \
\\<find-cmd-widget-field-keymap>\\[find-cmd-widget-default-action] in
\"*Widget Find*\" buffers."
  (interactive)
  (let ((grep-regexp (widget-value find-cmd-widget-grep-widget)))
    (if (equal "" grep-regexp)
        (find-cmd-widget-find-dired-action)
      (find-cmd-widget-find-grep-action grep-regexp))))

(defvar find-cmd-widget-field-keymap
  (let ((map (make-sparse-keymap)))
    (set-keymap-parent map widget-field-keymap)
    (define-key map [(control ?c) (control ?c)] #'find-cmd-widget-default-action)
    map))

(define-widget 'find-cmd-widget-editable-field 'editable-field
  "Doc..."
  :size 1
  :validate #'find-cmd-widget--just-true
  :value-get (lambda (w) (list (intern (widget-get w :menu-tag)) (widget-field-value-get w)))
  :keymap find-cmd-widget-field-keymap)

(defun find-cmd-widget-make-editable-field-widget (name &optional hint)
  (let ((name-string (symbol-name name)))
    `(find-cmd-widget-editable-field
      :menu-tag ,name-string
      :format ,(concat name-string ": %v" (and hint (format "  (%s)" hint)) "\n"))))

(defun find-cmd-widget-make-lname-widget (name &optional hint)
  (let ((name-string (symbol-name name)))
    `(find-cmd-widget-editable-field
      :menu-tag ,name-string
      :format ,(concat name-string ": %v" (and hint (format "  (%s)" hint)) "\n")
      :value-get
      ,(lambda (w)
         (when (eq 'L (car find-cmd-widget-current-global-option))
           (push (format "Global option `L` and `%s` used together" name-string)
                 find-cmd-widget-warnings))
         (list name (widget-field-value-get w))))))

(define-widget 'find-cmd-widget-n 'find-cmd-widget-editable-field
  "Widget to input find(1) n option arguments.")

(defun find-cmd-widget-make-n-widget (tag &optional hint)
  (when (symbolp tag) (cl-callf symbol-name tag))
  `(find-cmd-widget-n
    :menu-tag ,tag
    :format ,(concat tag ": %v" (and hint (format "  (%s)" hint)) "\n")))

(defun find-cmd-widget-make-signed-n-widget (name hint unit &optional default-sign)
  (find-cmd-widget--warn-arg-not-implemented hint)
  (let ((sname (symbol-name name)))
    `(group
      :tag ,sname
      :format ,(format "%s:\n%%v" (nth 3 (assoc name find-cmd-widget-find-arg-alist)))
      :value-get ,(lambda (w) `(,name ,(apply #'concat (widget-editable-list-value-get w))))
      :args
      ((find-cmd-widget-sign
        ,@(and default-sign `(:explicit-choice ,(cdr (assoc default-sign find-cmd-widget-sign-args)))))
       (find-cmd-widget-n
        :value-get ,#'widget-field-value-get
        :menu-tag ,unit
        :format ,(format "%%v %s\n" unit))))))

(defun find-cmd-widget-make-time-days-widget (name hint)
  (find-cmd-widget-make-signed-n-widget name hint "days" '<))

(defun find-cmd-widget-make-time-minutes-widget (name hint)
  (find-cmd-widget-make-signed-n-widget name hint "minutes" '<))

(defun find-cmd-widget-make-size-widget (name _hint)
  (let ((sname (symbol-name name)))
    `(group
      :tag ,sname
      :format ,(format "%s\n%%v" "File size is")
      :value-get
      ,(lambda (w)
         (let* ((values (widget-editable-list-value-get w))
                (value-string (apply #'concat values)))
           (pcase values
             (`(,(pred (equal "-")) ,(pred (equal "1")) . ,_)
              (push (format "size %s will only match empty files - see the find(1) doc"
                            value-string)
                    find-cmd-widget-warnings)))
           `(,name ,value-string)))
      :args
      ((find-cmd-widget-sign)
       (find-cmd-widget-n
        :value-get ,#'widget-field-value-get
        :format "%v ")
       ,(let ((args (mapcar (pcase-lambda (`(,descr ,val))
                              `(choice-item :tag ,descr
                                            :value-inline ,(lambda (_w) `(,val))
                                            :format "%[%t%] "))
                            '(("bytes"           "c")
                              ("KiB"             "k")
                              ("MiB"             "M")
                              ("GiB"             "G")
                              ("two-byte words"  "w")
                              ("512-byte blocks" "b")))))
          `(menu-choice
            :format "%vrounded up\n"
            :tag "unit"
            :explicit-choice ,(car args)
            :action ,(find-cmd-widget-menu-make-choice-action args)
            :args ,args))))))

(defun find-cmd-widget-make-type-widget (name hint)
  (find-cmd-widget--warn-arg-not-implemented hint)
  (cl-flet ((vget (s) (lambda (_w) (list (list s)))))
    `(checklist :tag ,(symbol-name name)
                :format "file type:\n%v"
                :value-get ;; validate here since using :validate is tricky
                ,(lambda (w)
                   (let ((v (widget-checklist-value-get w)))
                     (when (and (eq name 'type)
                                (eq 'L (car find-cmd-widget-current-global-option))
                                (seq-some (lambda (x) (string= (car x) "l")) v))
                       (push "Global option `L` and `type l` (\"symbolic link\") used together"
                             find-cmd-widget-warnings))
                     `(,name ,(mapconcat #'car v ","))))
                :args ((choice-item :tag "regular file"                   :value-inline ,(vget "f"))
                       (choice-item :tag "directory"                      :value-inline ,(vget "d"))
                       (choice-item :tag "symbolic link"                  :value-inline ,(vget "l"))
                       (choice-item :tag "block (buffered) special"       :value-inline ,(vget "b"))
                       (choice-item :tag "character (unbuffered) special" :value-inline ,(vget "c"))
                       (choice-item :tag "named pipe (FIFO)"              :value-inline ,(vget "p"))
                       (choice-item :tag "socked"                         :value-inline ,(vget "s"))
                       (choice-item :tag "door (Solaris)"                 :value-inline ,(vget "D"))))))

(defun find-cmd-widget-make-combiner-widget (combiner hint &optional with-operator)
  ;; With WITH-OPERATOR non-nil, return a compound widget that starts
  ;; with a `and' named button that can be switched to `or' without
  ;; effecting the specified children (the resulting value is effected
  ;; or course)
  (find-cmd-widget--warn-arg-not-implemented hint)
  (let ((sname (symbol-name combiner)))
    `(,(if with-operator 'find-expr-w/combiner 'find-expr)
      :tag ,sname
      :format ,(concat sname "\n%v")
      :value-inline ,(lambda (w)
                       `((,combiner ,@(if (not with-operator) (widget-child-value-get w)
                                        `(,(let ((v (widget-value w))) (cons (car v) (cadr v)))))))))))

(defun find-cmd-widget-make-combiner-widget-w/operator (combiner hint)
  (find-cmd-widget-make-combiner-widget combiner hint 'with-operator))

(define-widget 'find-expr-w/combiner 'group
  "Like `find-expr` but expressions can be `or'd with a toggle (`and` is implicit)."
  :value-inline (lambda (w)
                  (let ((w (widget-editable-list-value-get w)))
                    (if (eq (car w) 'and) (list (cadr w))
                      `(((,(car w) ,@(cadr w)))))))
  :args '((menu-choice
           :format "(`%v'ed)\n" ;maybe show only once multiple entries specified (-> :insert-before)?
           :args ((choice-item :tag "and" :value-inline (lambda (_) '(and))
                               :format "%[%t%]")
                  (choice-item :tag "or"  :value-inline (lambda (_) '(or))
                               :format "%[%t%]")))
          (find-expr :format "%v")))

(define-widget 'find-cmd-widget-file-widget 'item
  "Doc..."
  :action (lambda (w &rest _)
            (widget-value-set w (funcall (widget-get w :read-file-name-fun) w))
            (widget-apply w :notify w))
  :read-file-name-fun (lambda (w) (read-file-name "File: " nil (widget-value-value-get w)))
  :value "")

(defun find-cmd-widget-make-xnewer-widget (name hint)
  `(find-cmd-widget-file-widget
    :value-get ,(lambda (w) `(,name ,(expand-file-name (widget-value-value-get w))))
    :tag ,(symbol-name name)
    :format ,(concat (symbol-name name)
                     " than %[file%]: %v"
                     (when hint (format "          [%s]" hint))
                     "\n")))

(defun find-cmd-widget-make-perm-widget (name hint)
  (find-cmd-widget--warn-arg-not-implemented hint)
  (let ((sname (symbol-name name)))
    `(group
      :tag ,sname
      :format "file permissions are %v"
      :value-get ,(lambda (w) `(,name ,(apply #'concat (widget-editable-list-value-get w))))
      :args
      (,(let ((args (mapcar (pcase-lambda (`(,descr ,val))
                              `(choice-item :tag ,descr
                                            :value-inline ,(lambda (_w) `(,val))
                                            :format "%[%t%] "))
                            '(("exactly" "")
                              ("all set" "-")
                              ("any set" "/")))))
          `(menu-choice
            :format "%v:\n"
            :explicit-choice ,(cadr args)
            :action ,(find-cmd-widget-menu-make-choice-action args)
            :args ,args))
       (find-cmd-widget-editable-field
        :value-get ,#'widget-field-value-get :format "%v\n")))))

(define-widget 'find-cmd-widget-newerXY-X-widget 'menu-choice
  "Doc..."
  :args (delq nil
              (mapcar (pcase-lambda (`(,descr ,val))
                        (if (equal val "t") nil
                          `(choice-item :tag ,descr
                                        :format "%[%t%] "
                                        :value-get (lambda (_w) ,val))))
                      find-cmd-widget-newerXY-X-list)))

(define-widget 'find-cmd-widget-newerXY-Y-widget 'menu-choice
  "Doc..."
  :args (mapcar (pcase-lambda (`(,descr ,val))
                  `(choice-item :tag ,descr
                                :format "%[%t%] "
                                :value-get (lambda (_w) ,val)))
                find-cmd-widget-newerXY-X-list))

(let ((read-time (lambda (_w) (read-string "Reference time: ")))
      (ref-widget-default-format "%[like in date -d:%] %v\n")
      (ref-widget-defaul-help-echo "\
Mostly free format human readable date string such as \"Sun, 29
Feb 2004 16:21:42 -0800\" or \"2004-02-29 16:21:42\" or even
\"last Thursday\", \"yesterday\" or \"2 hours ago\".  An empty
string indicates the beginning of the day.  Full description in
the info documentation of GNU date."))
  (define-widget 'find-cmd-widget-newerXY-widget 'group
    "Doc..."
    :value-get (lambda (w) (pcase-let ((`(,x ,_ ,y ,ref) (widget-editable-list-value-get w)))
                             `(,(intern (concat "newer" x y)) ,ref)))
    :format "%v"
    :args `((find-cmd-widget-newerXY-X-widget :format "%v")
            (item :value "of the file" :format "%t ")
            (find-cmd-widget-newerXY-Y-widget
             :format "newer than %v"
             :notify ,(let ((last-was-t-flag t))
                        (lambda (w child &optional event)
                          (let ((ref-widget (nth 3 (widget-get (widget-get w :parent) :children)))
                                (this-is-t-flag (equal (widget-value w) "t")))
                            (if this-is-t-flag
                                (progn
                                  (widget-put ref-widget :format ref-widget-default-format)
                                  (widget-put ref-widget :help-echo ref-widget-defaul-help-echo)
                                  (widget-put ref-widget :read-file-name-fun read-time))
                              (widget-put ref-widget :format "%[of file:%] %v\n")
                              (widget-put ref-widget :help-echo nil)
                              (widget-put ref-widget
                                          :read-file-name-fun
                                          (lambda (w)
                                            (minibuffer-with-setup-hook
                                                (lambda ()
                                                  (delete-minibuffer-contents)
                                                  (insert (widget-value-value-get w)))
                                              (read-file-name "Reference file: ")))))
                            (widget-value-set
                             ref-widget
                             (if (eq this-is-t-flag last-was-t-flag) (widget-value ref-widget) ""))
                            (setq last-was-t-flag this-is-t-flag))
                          (widget-default-notify w child event))))
            (find-cmd-widget-file-widget
             :format ,ref-widget-default-format
             :read-file-name-fun ,read-time))))

(defun find-cmd-widget-defs ()
  (append
   (mapcar
    (pcase-lambda (`(,name ,descr ,_def))
      `(,name ,#'find-cmd-widget-make-user-pattern-widget "Additional" ,descr))
    (append find-cmd-widget-user-patterns find-cmd-widget-additional-patterns))
   find-cmd-widget-find-arg-alist))

;;;###autoload
(defun find-cmd-widget-define-main-widget (starting-find-expr &optional starting-value)
  (define-widget 'find-expr 'lazy
    "Doc..."
    :format "...:\n%v" ;this is/ should be overwritten by the caller
    :type `(editable-list

            ;; :append-button-args '(:tag "+") ;; that doesn't change all occurrences

            ;; :entry-format "%i\n%d %v" ;; that breaks intendation; %n not
            ;; implemented in `widget-editable-list-entry-create' :-(

            :delete-at
            (lambda (w child)
              ;; prevent the original function to change the value of CHILD; this
              ;; messes with my handling of values
              (cl-letf (((gv-synthetic-place (widget-get child :value)
                                             (lambda (v) `(widget-put child :value ,v)))))
                (widget-editable-list-delete-at w child)))

            :args
            ((menu-choice
              :tag "expr"
              :explicit-choice ,(find-cmd-widget--get-arg
                                 (assoc starting-find-expr (find-cmd-widget-defs)))
              :value ,starting-value
              ;; I hope this is the correct way to compute the choice of selections at run-time:
              :convert-widget
              ,(lambda (widget)
                 (let ((all-defs (find-cmd-widget-defs)))
                   (widget-put widget :action (find-cmd-widget-menu-make-choice-action all-defs t))
                   (widget-put widget :args (mapcar #'find-cmd-widget--get-arg all-defs)))
                 widget))))))

(find-cmd-widget-define-main-widget 'iname)

(define-widget 'find-cmd-widget-global-option 'menu-choice
  "Doc..."
  :format "Symlink Handling: %v\n"
  :action (find-cmd-widget-menu-make-choice-action find-cmd-widget-find-global-option-arglist)
  :notify (lambda (w &rest args)
            (setq find-cmd-widget-current-global-option (widget-value w))
            (apply #'widget-default-notify w args))
  :args (cons `(choice-item :tag "default"
                            :value-get ,(lambda (_w) nil))
              (mapcar #'find-cmd-widget--get-arg
                      find-cmd-widget-find-global-option-arglist)))

(define-widget 'find-cmd-widget-global-options 'editable-list
  "Doc..."
  :format (propertize "Global Options:\n%v%i\n\n" 'help-echo "\
Options affecting the operation of other tests and actions")
  :action (find-cmd-widget-menu-make-choice-action find-cmd-widget-find-arg-global-options-arglist)
  :notify (lambda (w &rest args)
            (setq find-cmd-widget-current-global-options (widget-value w))
            (apply #'widget-default-notify w args))
  :args
  `((menu-choice
     :tag "expr"
     :action ,(find-cmd-widget-menu-make-choice-action
               find-cmd-widget-find-arg-global-options-arglist t)
     :args ,(mapcar #'find-cmd-widget--get-arg find-cmd-widget-find-arg-global-options-arglist)
     :explicit-choice ,(find-cmd-widget--get-arg
                       (assoc 'maxdepth find-cmd-widget-find-arg-global-options-arglist)))))

(define-widget 'find 'group
  "Doc..."
  :format "%v"
  :value-get (lambda (w)
               (setq find-cmd-widget-warnings '())
               ;; (widget-editable-list-value-get w) returns a value in the form
               ;; ("/path/to/Dir" "find" (PATH...) SHANDLING (GLOBAL-OPTION...) (PATTERN...))
               (let* ((v (cddr (widget-editable-list-value-get w)))
                      (second-last (last v 2)))
                 (setf (car second-last) (nconc (car second-last) (cadr second-last))
                       (cdr second-last) nil)
                 v))
  :notify (pcase-lambda ((app widget-value v) &rest _)
            (widget-value-set find-cmd-widget-preview-widget-1
                              (string-trim-right (pp-to-string (find-cmd-widget-create-find-cmd v))))
            (widget-value-set find-cmd-widget-preview-widget-2
                              (find-cmd-widget-create-find-call v))
            (widget-value-set find-cmd-widget-warning-widget
                              (if find-cmd-widget-warnings
                                  (concat "\n"
                                          (mapconcat
                                           (apply-partially #'concat "Warning: ")
                                           find-cmd-widget-warnings "\n"))
                                ""))
            (with-current-buffer (widget-field-buffer find-cmd-widget-preview-widget-1)
              (set-buffer-modified-p nil)))
  :args
  `((find-cmd-widget-file-widget
     :format "%[Directory%]: %v\n"
     :convert-widget
     ,(lambda (widget)
        (widget-put widget :value default-directory)
        (widget-put widget :args nil)
        widget)
     :read-file-name-fun ,(lambda (w) (read-directory-name "Dir: " (widget-value w) nil t))
     :notify ,(lambda (w &rest args)
                (setq-local default-directory (widget-value w))
                (apply #'widget-default-notify w args)))
    (find-cmd-widget-file-widget
     :format "%[Executable%]: %v\n"
     :convert-widget
     ,(lambda (widget)
        (widget-put widget :value find-program)
        (widget-put widget :args nil)
        widget)
     :read-file-name-fun ,(lambda (_w) (string-trim-right (read-shell-command "find-program: ")))
     :notify (lambda (w &rest args)
               (setq-local find-program (widget-value w))
               (apply #'widget-default-notify w args)))
    (editable-list
     :value-create
     ;; Is there a better approach, Mauro?
     ,(lambda (w)
        (unless (widget-get w :value)
          (widget-put w :value (list default-directory)))
        (widget-editable-list-value-create w))
     :format "Paths:\n%v%i\n\n"
     :args
     ((find-cmd-widget-file-widget
       :format "find %[path%]: %v\n"
       ;; initialize VALUE dynamically:
       :convert-widget
       ,(lambda (widget)
          (widget-put widget :value default-directory)
          (widget-put widget :args nil)
          widget)
       :read-file-name-fun ,(lambda (w) (read-directory-name "Dir: " (widget-value w))))))
    (find-cmd-widget-global-option)
    (find-cmd-widget-global-options)
    (find-expr-w/combiner :format "Options: %v")))

(defvar find-cmd-widget-minor-mode-map
  (let ((map (make-sparse-keymap)))
    (set-keymap-parent map widget-keymap)
    (define-key map [(control ?c) (control ?c)] #'find-cmd-widget-default-action)
    (define-key map [?q] #'quit-window)
    (define-key map [remap self-insert-command] 'Custom-no-edit)
    (define-key map "\^m" #'Custom-newline)
    map))

(defvar find-cmd-widget-default-start-fun
  (lambda ()
    (setq-local find-cmd-widget-main-widget (widget-create 'find))))

(define-derived-mode find-cmd-widget-minor-mode nil "Find-Cmd-Wdgt"
  ;; Inheriting from no mode is what Custom-mode does; it has the
  ;; advantage that fields are correctly editable while still q can bury
  ;; the buffer when not in an editable field
  "Doc..."
  (use-local-map find-cmd-widget-minor-mode-map))

(put 'find-cmd-widget-minor-mode 'mode-class 'special)


;;;; Main commands

;;;###autoload
(defun find-cmd-widget (&optional start-fun)
  "Doc..."
  (interactive)
  (let ((buf (generate-new-buffer "*Widget Find*")))
    (pop-to-buffer buf)
    (kill-all-local-variables)
    (let ((inhibit-read-only t))
      (erase-buffer))
    (remove-overlays)
    (find-cmd-widget-minor-mode)
    (custom--initialize-widget-variables)
    (insert "Graphical interface for find(1) ")
    (widget-create 'push-button
                   :tag "manpage" :format "(%[%v%])"
	           :help-echo "Read the find(1) manpage"
                   :action (lambda (&rest _) (man "find(1)")))
    (insert "\n\n")
    (funcall (or start-fun find-cmd-widget-default-start-fun))
    (setq-local find-cmd-widget-warning-widget
                ;; FIXME: Avoid inserting a newline even when the value is the empty string
                (widget-create 'find-cmd-preview
                               :value ""
                               :value-face 'error))
    (let ((arrow ?⇩)
          (insert-space-before-copy-button
           (lambda () (insert (make-string 3 ?\s)))))
      (when (char-displayable-p arrow)
        (insert (propertize (concat "    " (string arrow)) 'face '(:height 2.0)) "\n"))
      (setq-local find-cmd-widget-preview-widget-1
                  (widget-create 'find-cmd-preview
                                 :value "\
Here a preview of the constructed s-exp will appear"))
      (funcall insert-space-before-copy-button)
      (widget-create 'push-button
                     :tag "Copy"
                     :help-echo "Copy `find-cmd' call to kill ring"
                     :action (lambda (&rest _)
                               (let ((s (widget-value find-cmd-widget-preview-widget-1)))
                                 (kill-new s)
                                 (message "%s" s))))
      (insert "\n")
      (when (char-displayable-p arrow)
        (insert (propertize (concat "    " (string arrow)) 'face '(:height 2.0)) "\n"))
      (setq-local find-cmd-widget-preview-widget-2
                  (widget-create 'find-cmd-preview :value "\
Here a preview of the constructed find(1) call will appear"))
      (funcall insert-space-before-copy-button)
      (widget-create 'push-button
                     :tag "Copy"
                     :help-echo "Copy \"find\" call to kill ring"
                     :action (lambda (&rest _)
                               (let ((s (widget-value find-cmd-widget-preview-widget-2)))
                                 (kill-new s)
                                 (message "%s" s)))))
    (insert "\n\nActions:\n")
    (widget-create 'push-button
                   :tag "find"
	           :help-echo "Run constructed find command"
                   :action #'find-cmd-widget-find-action)
    (insert "  ")
    (widget-create 'push-button
                   :tag "find-dired"
	           :help-echo "Run constructed find command as find-dired"
                   :action #'find-cmd-widget-find-dired-action)
    (insert "  ")
    (letrec ((grep-widget
              (widget-create
               `(group
                 :tag "Grep" :format ,(concat (propertize ":" 'display "") "%v")
                 :value-get ,(lambda (w) (cadr (widget-editable-list-value-get w)))
                 :args ((push-button
                         :tag "grep"
	                 :help-echo "Run grep on file list returned by constructed find command"
                         :action ,(lambda (&rest _)
                                    (find-cmd-widget-find-grep-action (widget-value grep-widget))))
                        (find-cmd-widget-editable-field :format " with regexp %v"
                                                        :value-get ,#'widget-field-value-get))))))
      (setq find-cmd-widget-grep-widget grep-widget))
    (insert "\n")
    (widget-setup)
    (goto-char (point-min))
    (search-forward-regexp "^Options:")
    (forward-line +1)))

(provide 'find-cmd-widget)


;;; Local Variables:
;;; eval: (el-search-hi-lock-add-pattern '(and (parent `(defvar find-cmd-widget-find-arg-alist . ,_) 3) (pred listp) l (guard (not (nth 3 l)))) 'el-search-hi-lock-unobtrusive-1)
;;; End:

;;; find-cmd-widget.el ends here
